searchNodes=[{"ref":"Boundary.html","title":"Boundary","type":"module","doc":"Definition of boundaries within the application. A boundary is a named group of modules which can export some of its modules, and depend on other boundaries. Boundary definitions can be used in combination with Mix.Tasks.Compile.Boundary to restrain cross-module dependencies. For example, you can use boundaries to prevent invocations from the context layer (e.g. MySystem) to the UI layer (e.g. MySystemWeb). Quick example The following code defines boundaries for a typical Phoenix based project generated with mix phx.new. defmodule MySystem do use Boundary, deps: [], exports: [] # ... end defmodule MySystemWeb do use Boundary, deps: [MySystem], exports: [Endpoint] # ... end defmodule MySystem.Application do use Boundary, deps: [MySystem, MySystemWeb] # ... end These boundaries specify the allowed cross-boundary usage: Modules from MySystemWeb may use the MySystem module, but not other MySystem.* modules. MySystem.Application code may use MySystem, MySystemWeb, and MySystemWeb.Endpoint modules. Defining a boundary A boundary is defined via use Boundary expression in the top-level (aka root) module. For example, the context boundary named MySystem can be defined as follows: defmodule MySystem do use Boundary, opts # ... end Module classification Based on the existing definitions, modules are classified into boundaries. Each module can belong to at most one boundary. A module doesn&#39;t need to belong to a boundary, in which case we say that the module is unclassified. Boundary membership is determined from the module name. In the previous example, we defined a single boundary, called MySystem. This boundary will contain the root module (MySystem), as well as all modules whose name starts with MySystem.. In addition, it&#39;s possible to extract some of the modules from a boundary into its own boundary. For example: defmodule MySystem do use Boundary, opts end defmodule MySystem.Endpoint do use Boundary, opts end Here, modules from MySystem.Endpoint namespace are promoted into its own boundary. It&#39;s worth noting that MySystem.Endpoint is considered as a peer boundary of MySystem, not its child. At the moment, nesting of boundaries (defining internal boundaries within other boundaries) is not supported by this library. Protocol implementation Consider the following protocol implementation: defimpl String.Chars, for: MySchema, do: # ... This code will generate the module String.Chars.MySchema. Therefore, the module sits in a completely different &quot;namespace&quot;. In addition, the desired boundary of such module can vary from one case to another. In some cases, a protocol implementation might be a UI concern, while in others, it might be a domain concern. For these reasons, protocol implementations are treated in a special way. A protocol implementation is by default unclassified (it doesn&#39;t belong to any boundary). However, the boundary checker will not emit a warning for unclassified protocol implementations. In addition, you can manually classify the protocol implementation, as demonstrated in the following example: defimpl String.Chars, for: MySchema do use Boundary, classify_to: MySystem # ... end Here, we&#39;re manually assigning the module (String.Chars.MySchema) to the MySystem boundary. Notice that :classify_to option is only allowed for protocol implementations. Plain modules (the ones which don&#39;t implement a protocol) can&#39;t be manually classified. Exports Exports are boundary modules which can be used by modules from other boundaries. A boundary always exports its root module, and it may additionally export other modules, which can be configured with the :exports option: defmodule MySystem do use Boundary, exports: [User] end In this example, we&#39;re defining the MySystem boundary which exports the modules MySystem and MySystem.User. All other modules of this boundary are considered to be internal, and they may not be used by modules from other boundaries. Dependencies Each boundary may depend on other boundaries. These dependencies are used to defined allowed cross-boundary module usage. A module from another boundary may only be used if: The callee boundary is a direct dependency of the caller boundary. The callee boundary exports the used module. For example: defmodule MySystem do use Boundary, exports: [User], deps: [] end defmodule MySystemWeb do use Boundary, exports: [], deps: [MySystem] end In this example we specify the following rules: Code from the MySystem boundary can&#39;t use any module from other boundaries. Code from the MySystemWeb boundary may use exports of the MySystem boundary (MySystem and MySystem.User). Of course, in-boundary cross-module dependencies are always allowed (any module may use all other modules from the same boundary). Ignored boundaries It is possible to exclude some modules from cross-boundary checks by defining an ignored boundary: defmodule MySystem do use Boundary, ignore?: true end When a boundary is ignored, all modules belonging to it can use any other module, and can be used by any other module. The purpose of this option is to support relaxing rules in some parts of your code. For example, you may wish to ignore boundary constraints for your test support modules. By introducing a top-level boundary for such modules (e.g. MySystemTest), and marking this boundary as ignored, you can easily achieve that. Another scenario is when introducing boundaries in an existing, possibly large project, which has many complex dependencies that can&#39;t be untangled trivially. In such case, ignored boundaries provide a mechanism for gradually introducing boundaries into the project. For example, you could first define ignored boundaries which encompass the entire system: defmodule MySystem do use Boundary, ignore?: true end defmodule MySystemWeb do use Boundary, ignore?: true end Now, you can pick smaller parts of your code where you can clean up the dependencies: defmodule MySystem.Context1 do use Boundary, exports: [...], deps: [] end defmodule MySystemWeb.Controller1 do use Boundary, exports: [], deps: [MySystem.Context1] end Going further, you can gradually expand the parts of your code covered by non-ignored boundaries. Once you&#39;re properly covering the entire system, you can remove the intermediate finer-grained boundaries, and specify the rules at the higher-level: defmodule MySystem do use Boundary, exports: [...], deps: [] end defmodule MySystemWeb do use Boundary, exports: [Endpoint], deps: [...] end"},{"ref":"Boundary.html#application/1","title":"Boundary.application/1","type":"function","doc":"Returns the boundary-specific view of the given application."},{"ref":"Boundary.html#t:application/0","title":"Boundary.application/0","type":"type","doc":""},{"ref":"Boundary.html#t:definition/0","title":"Boundary.definition/0","type":"type","doc":""},{"ref":"Boundary.html#t:name/0","title":"Boundary.name/0","type":"type","doc":""},{"ref":"Boundary.DefinitionError.html","title":"Boundary.DefinitionError","type":"exception","doc":""},{"ref":"Mix.Tasks.Compile.Boundary.html","title":"Mix.Tasks.Compile.Boundary","type":"task","doc":"Verifies cross-module function calls according to defined boundaries. This compiler reports all cross-boundary function calls which are not permitted, according to the current definition of boundaries. For details on defining boundaries, see the docs for the Boundary module. Usage Once you have configured the boundaries, you need to include the compiler in mix.exs: defmodule MySystem.MixProject do # ... def project do [ compilers: Mix.compilers() ++ [:boundary], # ... ] end # ... end When developing a library, it&#39;s advised to use this compiler only in :dev and :test environments: defmodule Boundary.MixProject do # ... def project do [ compilers: Mix.compilers() ++ extra_compilers(Mix.env()), # ... ] end # ... defp extra_compilers(:prod), do: [] defp extra_compilers(_env), do: [:boundaries] end Warnings Every invalid cross-boundary call is reported as a compiler warning. Consider the following example: defmodule MySystem.User do def auth() do MySystemWeb.Endpoint.url() end end Assuming that calls from MySystem to MySystemWeb are not allowed, you&#39;ll get the following warning: $ mix compile warning: forbidden call to MySystemWeb.Endpoint.url/0 (calls from MySystem to MySystemWeb are not allowed) (call originated from MySystem.User) lib/my_system/user.ex:3 Since the compiler emits warnings, mix compile will still succeed, and you can normally start your system, even if some boundary rules are violated. The compiler doesn&#39;t force you to immediately fix these violations, which is a deliberate decision made to avoid disrupting the development flow. At the same time, it&#39;s worth enforcing boundaries on the CI. This can easily be done by providing the --warnings-as-errors option to mix compile."},{"ref":"readme.html","title":"Boundary","type":"extras","doc":"Boundary Boundary is a library which helps managing and restraining cross-module dependencies in Elixir projects."},{"ref":"readme.html#status","title":"Boundary - Status","type":"extras","doc":"Highly experimental, untested, and unstable."},{"ref":"readme.html#documentation","title":"Boundary - Documentation","type":"extras","doc":"For a detailed reference see moduledoc here and here."},{"ref":"readme.html#basic-usage","title":"Boundary - Basic usage","type":"extras","doc":"To use this library, you first need to define the boundaries of your project. A boundary is a named group of one or more modules. Each boundary exports some (but not all!) of its modules, and can depend on other boundaries. During compilation, the boundary compiler will find and report all cross-module function calls which are not permitted according to the boundary configuration. Example The following code defines boundaries for a typical Phoenix based project generated with mix phx.new. defmodule MySystem do use Boundary, deps: [], exports: [] # ... end defmodule MySystemWeb do use Boundary, deps: [MySystem], exports: [Endpoint] # ... end defmodule MySystem.Application do use Boundary, deps: [MySystem, MySystemWeb] # ... end The configuration above defines three boundaries: MySystem, MySystemWeb, and MySystem.Application. Boundary modules are determined automatically from the boundary name. For example, the MySystem boundary contains the MySystem module, as well as any module whose name starts with MySystem. (e.g. MySystem.User, MySystem.User.Schema, ...). The configuration specifies the following rules: Modules residing in the MySystemWeb boundary are allowed to invoke functions from modules exported by the MySystem boundary. Modules residing in the MySystem.Application namespace are allowed to invoke functions from modules exported by MySystem and MySystemWeb boundaries. All other cross-boundary calls are not permitted. Next, you need to add the mix compiler: defmodule MySystem.MixProject do use Mix.Project def project do [ compilers: [:phoenix, :gettext] ++ Mix.compilers() ++ [:boundary], # ... ] end # ... end Boundary rules are validated during compilation. For example, if we have the following code: defmodule MySystem.User do def auth do MySystemWeb.Endpoint.url() end end The compiler will emit a warning: $ mix compile warning: forbidden call to MySystemWeb.Endpoint.url/0 (calls from MySystem to MySystemWeb are not allowed) (call originated from MySystem.User) lib/my_system/user.ex:3 The complete working example is available here. Because boundary is implemented as a mix compiler, it integrates seamlessly with editors which can work with mix compiler. For example, in VS Code with Elixir LS:"},{"ref":"readme.html#troubleshooting","title":"Boundary - Troubleshooting","type":"extras","doc":"Boundary loads the application Boundary loads the applications because it needs to fetch module attributes. If you&#39;re loading your application in your code, for example in test. Treat the case when the application is already loaded as a success case. case Application.load(:my_app) do :ok -&gt; :ok {:error, {:already_loaded, :my_app}} -&gt; :ok end"},{"ref":"readme.html#known-issues","title":"Boundary - Known Issues","type":"extras","doc":"Boundary will report incorrect invalid calls when multiple modules are defined in the same file (which often happens with Protocol implementations) This is caused by an Elixir bug that will be fixed in 1.10 (but will also be solved by moving to compilation tracers in https://github.com/sasa1977/boundary/issues/7) More info in the Elixir bug report: https://github.com/elixir-lang/elixir/issues/9322 As a workaround do not define multiple modules in one file (especially modules that are not in the same boundary)"},{"ref":"readme.html#roadmap","title":"Boundary - Roadmap","type":"extras","doc":"[ ] support nested boundaries (defining internal boundaries within a boundary) [ ] validate calls to external deps (e.g. preventing Ecto usage from MySystemWeb, or Plug usage from MySystem) [ ] support Erlang modules"},{"ref":"readme.html#license","title":"Boundary - License","type":"extras","doc":"MIT"}]